#ifndef MLN_KERNEL_CONTEXT_HPP
# define MLN_KERNEL_CONTEXT_HPP

# include <tuple>
# include <mln/core/image/image.hpp>
# include <mln/core/neighborhood/neighborhood.hpp>
# include <mln/accu/accumulator.hpp>
# include <mln/core/internal/intseq.hpp>
# include <boost/proto/proto.hpp>
# include <mln/kernel/types.hpp>
# include <mln/kernel/aggregate.hpp>
# include <mln/kernel/intro.hpp>

namespace mln
{

  namespace kernel
  {
    namespace proto = boost::proto;


    template <class FullExpr,
              int NumberOfReduction,
              class Neighborhood,
              class... I>
    struct kernel_context;


    /******************************************/
    /****          Implementation          ****/
    /******************************************/

    namespace internal
    {
      struct nil {};

      template <class... T>
      bool first_non_nil(bool v, T...)
      {
        return v;
      }

      template <class... T>
      bool first_non_nil(nil, T... x)
      {
        return first_non_nil(x...);
      }

    };


    template <class FullExpr,
              int NumberOfReduction,
              class Nbh,
              class...>
    struct kernel_context_;

    template <class FullExpr,
              class Nbh,
              int... K0, // 0...NumberOfReduction - 1
              int... K,  // 0...NumberOfImage - 1
              class... I>
    struct kernel_context_<FullExpr, Nbh, intseq<K0...>, intseq<K...>, I...>
    {
      struct fake_iter
      {
        template <class... T>
        fake_iter(T...) {}

        void init() const {}
        void next() const {}

        constexpr
        internal::nil
        finished() const { return internal::nil(); }
      };


      kernel_context_(const Nbh& nbh, I&&... images)
        : m_nbh(nbh),
          m_ima(std::forward<I>(images)...),
          m_px(rng::iter(m_ima.pixels())),
          m_spx(unzip_proxy_pixel_iterator<zip_iterator_t, K>(m_px)...),
          m_nx(rng::iter(m_nbh(std::get<K>(m_spx)))...)
      {
      }


      void p_init()
      {
        m_px.init();
      }

      void p_next()
      {
        m_px.next();
      }

      bool p_finished() const
      {
        return m_px.finished();
      }

      void n_init()
      {
        dontcare((std::get<K>(m_nx).init(), 1)... );
      }

      void n_next()
      {
        dontcare((std::get<K>(m_nx).next(), 1)... );
      }

      bool n_finished() const
      {
        return internal::first_non_nil(std::get<K>(m_nx).finished()...);
      }

      // Context operation.
      typedef std::tuple<I...> image_tuple_t;


      template<class Expr,
	       class Tag = typename proto::tag_of<Expr>::type,
	       class Args0 = typename Expr::proto_child0 >
      struct eval
	: proto::default_eval<Expr, kernel_context_, Tag>
      {};


      // Specialization for f_k(p)
      template <class Expr, int k>
      struct eval<Expr, proto::tag::terminal, ima_expr_tag<k, kernel::point>  >
      {
	typedef typename std::tuple_element<k, image_tuple_t>::type image_t;
	typedef mln_reference(image_t) result_type;

	result_type
	operator() (Expr&, kernel_context_& ctx) const
	{
	  return std::get<k>(ctx.m_spx)->val();
	}
      };

      // Specialization for f_k(n)
      template <class Expr, int k>
      struct eval<Expr, proto::tag::terminal, ima_expr_tag<k, kernel::neighbor>  >
      {
	typedef typename std::tuple_element<k, image_tuple_t>::type image_t;
	typedef mln_reference(image_t) result_type;

	result_type
	operator() (Expr&, kernel_context_& ctx) const
	{
	  return std::get<k>(ctx.m_nx)->val();
	}
      };

      // Specialization for A_k(...) pre-computed
      template <class Expr, int k>
      struct eval<Expr, proto::tag::terminal, aggregate_pl<k>  >
      {
	typedef typename std::tuple_element<k, accumulator_tuple_t>::type A;
	typedef typename mln::accu::result_of<A>::type result_type;

	result_type
	operator() (Expr&, kernel_context_& ctx) const
	{
	  return std::get<k>(ctx.m_accumulators).to_result();
	}
      };

      // Specialization for A(...) in computation
      template <class Expr, class Acc, class Arg0>
      struct eval<Expr, kernel::tag::aggregate<Acc>, Args0>
      {
	typedef void result_type;

	template <typename Accu>
	void
	operator() (Expr&, kernel_context_& ctx, Accu& aux) const
	{
	  aux.take( proto::eval(proto::child(expr), ctx) );
	}
      };

      // Specialization for Aggregation e.g Sum(f(n))
      // template <class Expr, class Acc, class Args0>
      // struct eval<Expr, kernel::tag::aggregate<Acc>, Args0>
      // {
      // 	typedef typename std::decay<typename proto::result_of::eval<Args0, kernel_context_>::type>::type T;
      // 	typedef typename accu::result_of<Acc, T>::type result_type;

      // 	result_type
      // 	operator() (Expr& expr, kernel_context_& ctx) const
      // 	{
      // 	  auto acc = accu::make_accumulator<Acc, T>( Acc() );
      // 	  acc.init();

      //     auto m_nx = ctx.m_nx; // Cache the neighbors iterators


      //     while (! internal::first_non_nil(std::get<K>(m_nx).finished()...))
      //       {
      //         acc.take( proto::eval(proto::child(expr), ctx) );
      //         dontcare( (std::get<K>(m_nx).next(), true)... );
      //       }
      // 	  return acc.to_result();
      // 	}
      // };


    private:
      typedef expression_traits<FullExpr> ETraits;
      typedef std::tuple<typename ETraits::template image<K>::with_n ...> N_used;


      typedef zip_image<I...>                                            image_t;
      typedef mln_pixiterator(image_t)                                   zip_iterator_t;
      typedef std::tuple<unzip_proxy_pixel_iterator<zip_iterator_t, K>...>     px_tuple_t;
      typedef std::tuple<
        typename std::conditional<
          std::tuple_element<K, N_used>::type::value,
          typename range_iterator<
            typename std::result_of<Nbh(unzip_proxy_pixel_iterator<zip_iterator_t, K>&)>::type
            >::type,
          fake_iter>::type
        ...
        >                                                                nx_tuple_t;


      Nbh                   m_nbh;
      image_t               m_ima;
      zip_iterator_t        m_px;
      px_tuple_t            m_spx;
      nx_tuple_t            m_nx;


      // For subexpr accumulator
      typedef std::tuple< 

    };

    template <class FullExpr, class Nbh, class... I>
    struct kernel_context
      : kernel_context_<FullExpr, Nbh, typename int_list_seq<sizeof...(I)>::type, I...>
    {
      typedef kernel_context_<FullExpr, Nbh, typename int_list_seq<sizeof...(I)>::type, I...> base;

      kernel_context(const Nbh& nbh, I&&... images)
        : base(nbh, std::forward<I>(images)...)
      {
      }
    };



  } // end of namespace mln::kernel

} // end of namespace mln

#endif //!MLN_KERNEL_CONTEXT_HPP
